/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "transaction.hh"
#include "buffer.hh"
#include "checksum_crc32.hh"
#include "driver_api.hh"
#include "seg_pool_mgr.hh"
#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <algorithm>
#include <cstring>

#include "dbase.hh"

using namespace sends;
using namespace std;

//========================================================================
//
//  Notes:
//  A) Clearing the transaction block
//      clear()
//          reset() - reset current transfer.
//              clear channels() - clear channel list
//                  release_segments() - release reserved segments
//                  clear channel list.
//                  reset default channel type
//              zero block number
//              reset write state.
//          disconnect socket
//
//========================================================================


//======================================  Return to transaction free queue
transaction::transaction(tran_queue& retQ, auth_server::auth_type at)
    : trans_id(0), home(&retQ), trans_sock(0), trans_auth(at), 
      trans_channels(1024), trans_state(ws_initial), trans_index_crc(0),
      trans_blockid(0), stage(0)
{
    trans_channels.clear();
}

//======================================  Return to transaction free queue
transaction::~transaction(void) {
    clear();
}

//======================================  Add a channel to the channel list
void
transaction::add_channel(chan_index inx) {
    trans_channels.push_back(inx);
    if (!Channel_List[inx].test_access(trans_auth.refUser().access())) {
	trans_channels.back().set_stat(-2);
    }
}

//======================================  Authenticate the user
void 
transaction::authenticate(buffer& b) {
    trans_auth.authenticate(*trans_sock, b);
}

//======================================  Return all transaction resources.
void
transaction::clear(void) {
    reset();
    trans_auth.close();
    set_socket(0);
}

//======================================  Return all transaction resources.
void
transaction::clear_channels(void) {
    release_segments();
    index_type N = trans_channels.size();
    for (index_type i=0; i<N; ++i) {
	driver_api* d = chan(i).driver();
	if (d) d->release_channel(chan(i).index());
	chan(i).set_driver(0);
    }
    trans_channels.clear();
    trans_chant = chan_type::kUnknown;
}

//======================================  Dump out the transaction descriptor
ostream& 
transaction::dump(ostream& out) const {
    out << "  Transaction (ID " << ident() << "):" << endl;
    if (trans_sock) {
        socket_api& s(*trans_sock);
        out << "  Socket:       id(" << s.socket_id() << ")" << endl;
        out << "                peer(" << s.peer_name() << ")" << endl;
        if (s.error()) {
           out << "                error " << s.error() << " at " 
	       << s.where() << ":" << strerror(s.sys_errno()) << endl;
        }
    } else {
        out << "  Not Connected!" << endl;
    }
    out << "  User:         " << ref_user().user() << endl;
    out << "  Command:      " << getCommandString(trans_code) << endl;
    if (trans_start || trans_stop || trans_stride) {
        out << "  Times:        " << trans_start << " -> " << trans_stop 
	    << " stride: " << trans_stride << endl;
    }
    if (!trans_epoch.all()) out << "  Epoch:        " << trans_epoch << endl;
    out << "  Stage:        " << stage << endl;
    out << "  Write state:  " << write_state_string(trans_state) << endl;

    //----------------------------------  List channels
    int N = trans_channels.size();
    if (N) out << "  Channel List (# entries: " << N << ")" << endl;
    else   out << "  No channels requested for this transaction" << endl;
    for (int i=0; i<N; ++i) {
	chan_index inx = trans_channels[i].index();
	out << "    " << Channel_List[inx].name() 
	    << " (index " << inx << ") fill: " << trans_channels[i].fill() 
	    << endl;
    }

    //----------------------------------  Reserved segments
    N = trans_segments.size();
    if (N) {
	out << "  Segment List (# entries: " << N << ")" << endl;
	segment::dump_header(out);
    } else {
	out << "  No segments reserved for this transaction!" << endl;
    }
    for (int i=0; i<N; i++) {
	seg_index jndx = trans_segments[i];
	out << "    ";
	Segment_List[jndx].dump(out); 
    }
    return out;
}

//======================================  Reply with error state
transaction::command_code 
transaction::getCommandCode(const std::string& str) {
#define TRY_COMMAND(x) if (str == getCommandString(x)) return x
    TRY_COMMAND(kNoop);
    TRY_COMMAND(kGetStatus);
    TRY_COMMAND(kGetData);
    TRY_COMMAND(kGetOnlineData);
    TRY_COMMAND(kGetServerStatus);
    TRY_COMMAND(kGetChannels);
    TRY_COMMAND(kCountChannels);
    TRY_COMMAND(kSegmentStatus);
    TRY_COMMAND(kGetSourceList);
    TRY_COMMAND(kSetEpoch);
    TRY_COMMAND(kGetSourceData);
    TRY_COMMAND(kChannelCRC);
    TRY_COMMAND(kServerProtocolVersion);
    TRY_COMMAND(kServerProtocolRevision);
    TRY_COMMAND(kListEpochs);
    TRY_COMMAND(kQuit);
#undef TRY_COMMAND
    throw runtime_error(string("Unknown command: ")+str);
}

//======================================  Reply with error state
const char*
transaction::getCommandString(command_code code) {
    switch (code) {
    case kNoop:
	return "noop";
    case kGetStatus:
	return "get-status";
    case kGetData:
	return "get-data";
    case kGetOnlineData:
	return "get-online-data";
    case kGetServerStatus:
	return "get-server-status";
    case kGetChannels:
	return "get-channels";
    case kCountChannels:
	return "count-channels";
    case kQuit:
	return "quit";
    case kSegmentStatus:
	return "segment-status";
    case kGetSourceList:
	return "get-source-list";
    case kSetEpoch:
	return "set-epoch";
    case kGetSourceData:
	return "get-source-data";
    case kChannelCRC:
	return "get-channel-crc";
    case kServerProtocolVersion:
	return "server-protocol-version";
    case kServerProtocolRevision:
	return "server-protocol-revision";
    case kListEpochs:
	return "list-epochs";
    default:
	break;
    }
    return "unknown";
}

//======================================  Read a line into a buffer.
bool 
transaction::next_stride(void) {
    trans_start += stride();
    bool more = true;
    if (trans_state == ws_end_data) {
	reset();
	more = false;
    } else if (trans_stop == 0 && trans_code == kGetOnlineData) {
	release_segments(trans_start);
    } else if (trans_start < trans_stop) {
	release_segments(trans_start);
    } else {
	reset();
	set_state(ws_end_data);
	more = false;
    }
    return more;
}

//======================================  Read a line into a buffer.
#define RESPONSE_LENGTH 4
int
transaction::put_response(int resp) {
    char rstr[RESPONSE_LENGTH];
    for (int i=RESPONSE_LENGTH; i--; ) {
	int x = resp & 15;
	if (x < 10) rstr[i] = '0' + x;
	else        rstr[i] = 'a' + (x-10);
	resp >>= 4;
    }
    return trans_sock->write(rstr, sizeof(rstr));
}

//======================================  Read a line into a buffer.
int
transaction::read(buffer& buf) {
    long maxlen = buf.capacity();
    long nRead = trans_sock->gets(buf.data(), maxlen);
    buf.push(nRead);
    return nRead;
}

//======================================  Return to transaction free queue
void
transaction::release(void) {
    clear();
    home->push(this);
}

//======================================  Release all segments
void 
transaction::release_segments(void) {
    int N=trans_segments.size();
    for (int i=0; i<N; ++i) {
	seg_index inx = trans_segments[i];
	trans_segments[i] = Segment_List.end();
	Segment_List[inx].unuse();
    }
    trans_segments.clear();
}

//======================================  Release segments before specified GPS
void 
transaction::release_segments(gps_type before) {
    int N=trans_segments.size();
    int j=0;
    for (int i=0; i<N; ++i) {
	seg_index inx = trans_segments[i];
	if (before < Segment_List[inx].getEndGPS()) {
	    if (i != j) trans_segments[j++] = inx;
	    else        j++;
	} else {
	    Segment_List.release(inx);
	}
    }
    trans_segments.erase(trans_segments.begin()+j, trans_segments.begin()+N);
}

//======================================  Reserve all available segments
void 
transaction::reserve_segments(void) {
    int N = trans_channels.size();
    for (int i=0; i<N; ++i) {
	chan_index inx = trans_channels[i].index();
	gps_type start = trans_channels[i].fill();
	if (start < trans_start) start = trans_start;
	gps_type stop  = getStrideEnd();
	while (start < stop) {
	    seg_index seg = Segment_List.reserve(inx, start);
	    if (Segment_List.is_null(seg)) break;
	    const_seg_iter segit = std::find(trans_segments.begin(),
					     trans_segments.end(),
					     seg);
	    if (segit == trans_segments.end()) trans_segments.push_back(seg);
	    else			       Segment_List[seg].unuse();
	    start = Segment_List[seg].getEndGPS();
	    trans_channels[i].set_fill(start);
	}
    }
}

//======================================  Reset command processing
void 
transaction::reset(void) {
    clear_channels();
    set_stage(0);
    trans_state     = ws_initial;
    trans_index_crc = 0;
    trans_blockid   = 0;
}

//======================================  Set the transaction ID
void
transaction::set_socket(socket_api* sock) {
    if (trans_sock && trans_sock != sock) {
	delete trans_sock;
	trans_sock = 0;
    }
    trans_sock = sock;
}

//======================================  Set the transaction epoch
void
transaction::set_epoch(const epoch& epic) {
    trans_epoch = epic;
}

//======================================  Set the transaction ID
void
transaction::set_id(long id) {
    trans_id = id;
}

//======================================  Set the transaction command
void 
transaction::setCommand(command_code cmd) {
    trans_code = cmd;
}

//======================================  Set the transaction command
void 
transaction::setCommand(const std::string& cmd) {
    trans_code = getCommandCode(cmd);
}

//======================================  Set the transaction command
void
transaction::setDefType(chan_type t) {
    trans_chant = t;
}

//======================================  Set the last stage ID
void 
transaction::set_stage(long s) {
    stage = s;
}

//======================================  Set the transaction write status
void
transaction::set_state(write_state s) {
    trans_state = s;
}

//======================================  Set the times
void 
transaction::setTime(gps_type start, gps_type stop, gps_type dt) {
    trans_start  = start;
    trans_stop   = stop;
    if (!dt) trans_stride = trans_stop - trans_start;
    else     trans_stride = dt;
}

//======================================  Set the fill gps for all channels
void 
transaction::start_fill(gps_type t) {
    index_type N = trans_channels.size();
    for (index_type i=0; i<N; ++i) {
	trans_channels[i].set_fill(t);
    }
}

//======================================  Test the index checksum
bool
transaction::test_index_crc(const std::string& hash) {
    checksum_crc32 chk;
    chk.add(hash);
    uint32_t chksum = chk.result();
    if (chksum == trans_index_crc) return false;
    trans_index_crc = chksum;
    return true;
}

//======================================  Test resource usage
bool
transaction::test_limits(void) const {
    //----------------------------------  Set up the buffer type counts.
    const index_type max_type(4);
    const double     max_buffer_frac(0.8);
    index_type buf_count[max_type];
    for (index_type i=0; i<max_type; i++) buf_count[i] = 0;

    //----------------------------------  Loop over channels, count buffers.
    index_type N = trans_channels.size();
    for (index_type i=0; i<N; ++i) {
	index_type ichn = trans_channels[i].index();
	seg_pool_mgr::select_type btyp = Buffer_Pool.get_buffer_class(ichn);
	if (btyp < 4) {
	    index_type bsize  = Buffer_Pool.size(btyp);
	    index_type nBytes = Channel_List[ichn].data_length(stride());
	    double     rbuf   = double(nBytes) / double(bsize);
	    int        nbuf   = int(rbuf);
	    if (rbuf > double(nbuf)) nbuf++;
	    buf_count[btyp] += nbuf + 1;
	}
    }

    //----------------------------------  Check availability for buffer types
    bool allow = true;
    for (index_type i=0; i<max_type; i++) {
	if (buf_count[i] == 0) continue;
	if (double(buf_count[i]) > max_buffer_frac*Buffer_Pool.buffer_count(i)){
	    allow = false;
	}
    }
    return allow;
}

std::string
transaction::write_state_string(write_state s) {
    string str;
    switch (s) {
    case ws_initial: 
	str = "Starting state, no response has been given";
	break;
    case ws_response:
	str = "Initial response (tid, flag) has been sent to the client.";
	break;
    case ws_chandata:
	str = "Channel data have been sent to the client.";
	break;
    case ws_end_data:
	str = "No more data available";
	break;
    default:
	str= "Undefined state";
    }
    return str;
 }
