/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "write_stage.hh"
#include "available.hh"
#include "ccstream.hh"
#include "dbase.hh"
#include "sysexcept.hh"
#include <iostream>
#include <sstream>
#include <netinet/in.h>
#include <signal.h>

typedef std::vector<std::string> str_vect;

using namespace sends;
using namespace std;

//======================================  Construct write stage
write_stage::write_stage(int nThread, int nTrans)
  : seda_stage(nThread, nTrans)
{
}

//======================================  Destroy write stage
write_stage::~write_stage(void) {
}

//======================================  Stage name.
const char*
write_stage::stage_name(void) const {
    return "write";
}

//======================================  Data transfer junk
static void
xfer_bytes(transaction* trans, const char* data, int len) {
    socket_api& s(trans->ref_socket());
    int rc = s.write(data, len);
    if (rc < 0) {
	string msg = "xfer_bytes: data write error in ";
	msg += s.where();
	if (s.error() != socket_api::serr_timeout) {
	    throw system_error(msg, s.sys_errno());
	}
	msg += ": time out";
	throw std::runtime_error(msg);
    }
}

static void
xfer_int(transaction* trans, int32_t data) {
    int32_t buf = htonl(data);
    xfer_bytes(trans, reinterpret_cast<char*>(&buf), sizeof(buf));
}

static void
xfer_float(transaction* trans, float data) {
    union {int32_t i; float f;} d;
    d.f = data;
    xfer_int(trans, d.i);
}

static void
xfer_string(transaction* trans, const std::string& str) {
    int len = str.size();
    xfer_int(trans, len);
    xfer_bytes(trans, str.c_str(), len);
}

static void
xfer_hex2(transaction* trans, unsigned short data) {
    char str[4];
    for (int i=0; i<4; i++) {
	int x = (data >> (12-i*4)) & 15;
	if (x < 10) str[i] = '0' + x;
	else        str[i] = 'a' + x - 10;
    }
    xfer_bytes(trans, str, 4);
}

static void
xfer_hex4(transaction* trans, unsigned int data) {
    xfer_hex2(trans, (unsigned short)((data >> 16) & 0xffff));
    xfer_hex2(trans, (unsigned short)(data         & 0xffff));
}

static void
xfer_segdata(transaction* trans, seg_index is, gps_type start, gps_type end) {
    const char* ipstart = Segment_List[is].refData(start);
    const char* ipend   = Segment_List[is].refData(end);
    long nBytes = ipend - ipstart;
    xfer_bytes(trans, ipstart, nBytes);
}

//======================================  Send data to the client.
void
write_stage::filter(transaction* trans) {
    trans->set_stage(500);
    try {
	switch (trans->command()) {

	//------------------------------  Don't do anything
	case transaction::kNoop:
	    trans->dump(cout);
	    break;

	//------------------------------  Send the server status information
	case transaction::kGetStatus:
	    write_status(trans);
	    break;

	//------------------------------  Send a chunk of online/offline data.
	case transaction::kGetData:
	case transaction::kGetOnlineData:
	    switch (trans->get_state()) {
	    case transaction::ws_initial:
		debug_msg(dlvl_detail, trans->ident(), "writing trans ID");
		xfer_hex4(trans, trans->ident());
		if (trans->command() == transaction::kGetOnlineData) {
		    debug_msg(dlvl_data, trans->ident(), "transfer type=0");
		    xfer_int(trans, 0); // "online" flag
		} else{
		    debug_msg(dlvl_data, trans->ident(), "transfer type=1");
		    xfer_int(trans, 1); // "offline" flag
		}
		trans->set_state(transaction::ws_response);
	    case transaction::ws_response:
	    case transaction::ws_chandata:
		write_status(trans); // write_status checks index hash
		write_data(trans);
	    case transaction::ws_end_data:
		break;
	    }
	    break;

	//------------------------------  Spew server status
	case transaction::kGetServerStatus:
	    write_server_status(trans);
	    break;

	//------------------------------  Get a list of channels
	case transaction::kGetChannels:
	    write_channel_list(trans);
	    break;

	//------------------------------  Get a count of channels
	case transaction::kCountChannels:
	    write_channel_count(trans);
	    break;

	//------------------------------  Get a crc of channels
	case transaction::kChannelCRC:
	    write_channel_crc(trans);
	    break;

	//------------------------------  Get a source list.
	case transaction::kGetSourceList:
	    write_source_list(trans, false);
	    break;

	//------------------------------  Get an available source data list
	case transaction::kGetSourceData:
	    write_source_list(trans, true);
	    break;

	//------------------------------  Print current segment status
	case transaction::kSegmentStatus:
	    Segment_List.dump_by_channel(cout);
	    break;

	//------------------------------  Output server protocol revision
	case transaction::kServerProtocolRevision:
	    if (debug(dlvl_detail)) {
		ostringstream msg;
		msg << "Writing revision: " << SERVER_PROTOCOL_REVISION
		    << std::endl;
		debug_msg(dlvl_detail, trans->ident(), msg.str());
	    }
	    xfer_int(trans, SERVER_PROTOCOL_REVISION);
	    break;

	//------------------------------  Output server protocol revision
	case transaction::kServerProtocolVersion:
	    if (debug(dlvl_detail)) {
		ostringstream msg;
		msg << "Writing version: " << SERVER_PROTOCOL_VERSION
		    << std::endl;
		debug_msg(dlvl_detail, trans->ident(), msg.str());
	    }
	    xfer_int(trans, SERVER_PROTOCOL_VERSION);
	    break;

	//------------------------------  Output server protocol revision
	case transaction::kListEpochs:
	    write_epochs(trans);
	    break;
	default:
	    break;
	};

	//------------------------------  Release data already transferred
	trans->set_stage(510);
	if (trans->next_stride()) put_alternate(trans); // Fetch more data
	else                      put_output(trans);    // Get another command
    } catch (exception& e) {
	trans->set_stage(520);
	reject(trans, e.what());
    }
}

struct nds2_chan_status {
    int status;
    int offset;
    int dat_type;
    float rate;
    float signal_offset;
    float signal_slope;
};

//======================================  Send a status reply
void
write_stage::write_status(transaction* trans) {
    int N = trans->getNChan();
    long offset = 0;
    gps_type start_gps = trans->getStartGPS();
    //gps_type end_gps = trans->getStrideEnd();

    //----------------------------------  build status hash string
    std::ostringstream hash;
    for (int i=0; i < N; ++i) {
	chan_index cinx = trans->chan(i).index();
	const chan_entry& ce(Channel_List[cinx]);
        long nBytes = 0;
	if (trans->chan(i).available()) {
	    nBytes = long(ce.rate() * ce.data_size() * trans->stride()); 
	    hash << nBytes;
	} else if (trans->chan(i).status() == -2) {
	    hash << -DAQD_ACCESS;
	} else {
	    hash << -DAQD_NOT_FOUND;
	}
	hash << " " << ce.type().code() << " " << ce.dtype().code() 
	     << " " << ce.rate();

	seg_index iseg = Segment_List.first_gt(cinx, start_gps);
	if (!Segment_List.is_null(iseg)) {
	    hash << " " << Segment_List[iseg].bias() 
		 << " " << Segment_List[iseg].scale();
	}
	offset += nBytes;
    }
    if (!trans->test_index_crc(hash.str())) return;

    //----------------------------------  build status structure list.
    std::vector<nds2_chan_status> statlist(N);
    offset = 0;
    long lTotal = 4*sizeof(int); // length of header exclusive of length word
    for (int i=0; i < N; ++i) {
	chan_index cinx = trans->chan(i).index();
	const chan_entry& ce(Channel_List[cinx]);
        long nBytes = 0;
	if (trans->chan(i).available()) {
	    nBytes = long(ce.rate() * ce.data_size() * trans->stride()); 
	    statlist[i].status    = nBytes;
	} else if (trans->chan(i).status() == -2) {
	    statlist[i].status    = -DAQD_ACCESS;
	} else {
	    statlist[i].status    = -DAQD_NOT_FOUND;
	}
	statlist[i].offset    = offset;
	statlist[i].dat_type  = (ce.type().code() << 16) + ce.dtype().code();
	statlist[i].rate      = ce.rate();

	seg_index iseg = Segment_List.first_gt(cinx, start_gps);
	if (Segment_List.is_null(iseg)) {
	    statlist[i].signal_offset = 0.0;
	    statlist[i].signal_slope  = 1.0;
	} else {
	    statlist[i].signal_offset = Segment_List[iseg].bias();
	    statlist[i].signal_slope  = Segment_List[iseg].scale();
	}
	offset += nBytes;
	lTotal += sizeof(nds2_chan_status);
    }
    if (!offset) {
	debug_msg(dlvl_detail, trans->ident(), "write_status: no channel data");
	return;
    }

    //----------------------------------  write block-id/chan-count
    //if (debug(2)) cout << "write_status: writing number of channels" << endl;
    //xfer_int(trans, N);

    //----------------------------------  write block-header
    debug_msg(dlvl_detail, trans->ident(), "write_status: writing block header");
    xfer_int(trans, lTotal);
    xfer_int(trans, -1);
    xfer_int(trans, trans->getStartGPS());
    xfer_int(trans, 0);
    xfer_int(trans, trans->block_seq());
   

    //----------------------------------  write block-id/chan-count
    debug_msg(dlvl_detail, trans->ident(), "write_status: writing channel status");
    for (int i=0; i < N; ++i) {
	xfer_int  (trans, statlist[i].status );
	xfer_int  (trans, statlist[i].offset);
	xfer_int  (trans, statlist[i].dat_type);
	xfer_float(trans, statlist[i].rate);
	xfer_float(trans, statlist[i].signal_offset);
	xfer_float(trans, statlist[i].signal_slope);
    }

    trans->set_state(transaction::ws_chandata);
}

//======================================  Send a data reply
void 
write_stage::write_data(transaction* trans) {
    //----------------------------------  Count the total bytes
    int N = trans->getNChan();
    long lTotal = 0;
    for (int i=0; i < N; ++i) {
	const chan_entry& ce(Channel_List[trans->chan(i).index()]);
	if (trans->chan(i).available()) {
	    lTotal += long(ce.rate() * ce.data_size() * trans->stride()); 
	}
    }

    //----------------------------------  Write the data block header.
    debug_msg(dlvl_detail, trans->ident(), "write_data: writing block header");
    xfer_int(trans, lTotal + 4*sizeof(int));
    xfer_int(trans, trans->stride());
    xfer_int(trans, trans->getStartGPS());
    xfer_int(trans, 0);
    xfer_int(trans, trans->block_seq());
    if (!lTotal) {
	debug_msg(dlvl_detail, trans->ident(), "write_data: no channel data");
	trans->set_state(transaction::ws_end_data);
	return;
    }

    //----------------------------------  get the data from the segments
    debug_msg(dlvl_detail, trans->ident(), "write_data: writing channel data");
    gps_type start_gps = trans->getStartGPS();
    gps_type   end_gps = trans->getStrideEnd();
    for (int i=0; i < N; ++i) {
	if (! trans->chan(i).available()) continue;

	//------------------------------  get stride limits, loop over segments
	chan_index  chan = trans->chan(i).index();
	gps_type start = start_gps;
	while (start < end_gps) {
	    seg_index iseg = Segment_List.find(chan, start, end_gps);
	    if (Segment_List.is_null(iseg)) {
	    	throw logic_error("Data not available");
	    }
	    gps_type seg_end = Segment_List[iseg].getEndGPS();
	    if (end_gps < seg_end) seg_end = end_gps;
	    xfer_segdata(trans, iseg, start, seg_end);
	    start = seg_end;
	}
    }
}

//======================================  Send a data reply
void 
write_stage::write_epochs(transaction* trans) {
    ostringstream out;
    Epoch_Table.list(out);
    xfer_string(trans, out.str());
}

//======================================  Send a server status reply
void 
write_stage::write_server_status(transaction* trans) {
}

//======================================  Test if a channel fits the descriptor
bool
write_stage::test_channel( chan_type ctype, gps_type gps, 
			   const chan_entry& ce) const 
{
    //----------------------------------  Check for legitimate type code
    if (ctype != chan_type::kUnknown && ctype != ce.type()) return false;

    //----------------------------------  gps == 0 means no gps test.
    if (!gps) return true;

    //----------------------------------  gps != 0 only for persistent data
    if (ce.type() == chan_type::kOnline) return false;

    //----------------------------------  Scan source list for available data.
    chan_index nSource = ce.getNSource();
    for (chan_index i=0; i < nSource; ++i) {
	const chan_entry::source_entry& src = ce.source(i);
	if (src.test_gps(gps) && Source_List.gps_valid(src.name(), gps)) {
	    return true;
	}
    }
    return false;
}

//======================================  Test if a channel fits the descriptor
bool
write_stage::test_channel( chan_type ctype, const epoch& epic, 
			   const chan_entry& ce) const 
{
    //----------------------------------  Epoch is ignored for online channels
    if (ctype == chan_type::kOnline && ctype == ce.type()) return true;

    //----------------------------------  Check availability for selected type
    if (ctype == chan_type::kUnknown || ctype == ce.type()) {
	chan_index nSource = ce.getNSource();
	for (chan_index i=0; i < nSource; ++i) {
	    if (ce.source(i).test_epoch(epic)) return true;
	}
    }
    return false;
}

//======================================  Build an index of channels

chan_index
write_stage::build_channel_index(const transaction& trans, 
				 index_vect& chanlist) const {
    chan_type ctype = trans.getDefType();
    gps_type  gps   = trans.getStartGPS();
    chan_index N    = Channel_List.size();
    chan_index nChan;

    //----------------------------------  Test for selected gps
    if (gps) {
	chanlist.reserve(N);
	for (chan_index i=0; i<N; ++i) {
	    if (test_channel(ctype, gps, Channel_List[i])) {
		chanlist.push_back(i);
	    }
	}
	nChan = chanlist.size();
    }

    //----------------------------------  Test for selected epoch
    else if (!trans.ref_epoch().all()) {
	chanlist.reserve(N);
	epoch epic = trans.ref_epoch();
	for (chan_index i=0; i<N; ++i) {
	    if (test_channel(ctype, epic, Channel_List[i])) {
		chanlist.push_back(i);
	    }
	}
	nChan = chanlist.size();
    }

    //----------------------------------  Test for selected channel type
    else if (ctype != chan_type::kUnknown) {
	chanlist.reserve(N);
	for (chan_index i=0; i<N; ++i) {
	    if (test_channel(ctype, gps, Channel_List[i])) {
		chanlist.push_back(i);
	    }
	}
	nChan = chanlist.size();
    }

    //----------------------------------  All channels are selected.
    else {
	nChan = N;
    }
    return nChan;
}

//======================================  Send a channel count
void 
write_stage::write_channel_count(transaction* trans) {
    debug_msg(dlvl_meta, trans->ident(), "Writing channel count");
    chan_type ctype = trans->getDefType();
    gps_type  gps   = trans->getStartGPS();
    chan_index N = Channel_List.size();
    chan_index nChan = 0;

    //----------------------------------  Test for selected gps
    if (gps) {
	for (chan_index i=0; i<N; ++i) {
	    if (test_channel(ctype, gps, Channel_List[i])) nChan++;
	}
    }

    //----------------------------------  Test for selected epoch
    else if (!trans->ref_epoch().all()) {
	epoch epic = trans->ref_epoch();
	for (chan_index i=0; i<N; ++i) {
	    if (test_channel(ctype, epic, Channel_List[i])) nChan++;
	}
    }

    //----------------------------------  Test for selected channel type
    else if (ctype != chan_type::kUnknown) {
	for (chan_index i=0; i<N; ++i) {
	    if (test_channel(ctype, gps, Channel_List[i])) nChan++;
	}
    }

    //----------------------------------  All channels qualify
    else {
	nChan = N;
    }

    //----------------------------------  Send the result.
    debug_msg(dlvl_detail, trans->ident(), "Send nChan");
    xfer_int(trans, nChan);
}

//======================================  Send a channel  list
void 
write_stage::write_channel_list(transaction* trans) {
    debug_msg(dlvl_meta, trans->ident(), "Writing channel list");

    //----------------------------------  build list of channel indices
    index_vect chanlist;
    chan_index N     = Channel_List.size();
    chan_index nChan = build_channel_index(*trans, chanlist);
    
    //----------------------------------  Send the result.
    debug_msg(dlvl_detail, trans->ident(), "Send nChan");
    xfer_int(trans, nChan);

    //----------------------------------  Send channel data.
    debug_msg(dlvl_detail, trans->ident(), "Send Channel data");
    for ( chan_index inx=0; inx<nChan; ++inx) {
	chan_index i = inx;
	if (nChan != N) i = chanlist[inx];
	ostringstream ostr;
	Channel_List[i].display(ostr) << ends;
	if (debug(dlvl_data)) {
	    string msg = "Send Channel:";
	    msg += ostr.str();
	    debug_msg(dlvl_data, trans->ident(), msg);
	}
	xfer_string(trans, ostr.str());
    }
}

//======================================  Send crc-32 of the channel list
void 
write_stage::write_channel_crc(transaction* trans) {
    debug_msg(dlvl_meta, trans->ident(), "Writing channel crc");
    index_vect chanlist;
    chan_index N     = Channel_List.size();

    //----------------------------------  Get the index
    chan_index nChan = build_channel_index(*trans, chanlist);

    //----------------------------------  Send channel data.
    ocrcstream ostr;
    debug_msg(dlvl_detail, trans->ident(), "Send Channel data");
    if (nChan == N) {
	for (chan_index inx=0; inx<nChan; ++inx) {
	    Channel_List[inx].display(ostr) << ends;
	}
    }
    else {
	for ( chan_index inx=0; inx<nChan; ++inx) {
	    Channel_List[chanlist[inx]].display(ostr) << ends;
	}
    }
    
    //----------------------------------  Send the result.
    debug_msg(dlvl_detail, trans->ident(), "Send Channel crc");
    ostr.flush();
    xfer_int(trans, ostr.get_crc32());
}

//======================================  Send a channel count
void 
write_stage::write_source_list(transaction* trans, bool avail) {
    debug_msg(dlvl_meta, trans->ident(), "Writing source list");
    gps_type t = trans->getStartGPS();

    //----------------------------------  Loop over requested channels
    ostringstream response;
    int N = trans->getNChan();
    for (int i=0; i < N; ++i) {
	if (i) response << " ";
	const chan_entry& ce(Channel_List[trans->chan(i).index()]);
	response << ce.name() << " {";

	//------------------------------  Full list?
	if (!t || avail) {
	    //--------------------------  Count the unique source types.
	    str_vect unique_src;
	    index_type nSrc = ce.getNSource();
	    for (index_type iSrc=0; iSrc<nSrc; iSrc++) {
		const chan_entry::source_entry& se = ce.source(iSrc);
		if (unique_src.empty()) {
		    unique_src.push_back(se.name());
		}
		else if (!std::count(unique_src.begin(), 
				     unique_src.end(),
				     se.name())) {
		    unique_src.push_back(se.name());
		}
	    }

	    //--------------------------  Loop over source types... Spew.
	    nSrc = unique_src.size();
	    for (index_type iSrc=0; iSrc<nSrc; iSrc++) {
		const string& src_i = unique_src[iSrc];

		//----------------------  Get the defined epochs
		available av;
		av.set_list(ce, src_i);
		av &= trans->ref_epoch();
		if (debug(dlvl_data)) {
		    ostringstream msg;
		    msg << "Channel " << ce.name() << " written: ";
		    av.src_list(src_i, msg);
		    debug_msg(dlvl_data, trans->ident(), msg.str());
		}

		//----------------------  fold in the data availability
		if (avail) {
		    available src;
		    const loc_list* ll=Source_List.list_ptr(loc_list::kCache,
							    src_i);
		    if (ll) src.set_list(*ll);
		    ll = Source_List.list_ptr(loc_list::kArchive, src_i);
		    if (ll) {
			available arch;
			arch.set_list(*ll);
			src |= arch;
		    }

		    if (debug(dlvl_data)) {
		        ostringstream msg;
		        msg << "Channel " << ce.name() << " source available: ";
		        src.src_list(src_i, msg);
		        debug_msg(dlvl_data, trans->ident(), msg.str());
		    }

		    av &= src;
		}
		if (iSrc) response << " ";
		av.coalesce();
		av.src_list(src_i, response);
	    }
	}

	//------------------------------  Loop over sources
	else {
	    bool ok = false;
	    index_type nSrc = ce.getNSource();
	    for (index_type iSrc=0; iSrc<nSrc; ++iSrc) {
		const chan_entry::source_entry& se = ce.source(iSrc);
		if (se.test_gps(t) && Source_List.gps_valid(se.name(), t)) {
		    if (ok) response << " ";
		    response << se.name() << ":";
		    se.refEpoch().display(response, epoch::st_trans);
		    ok = true;
		}
	    }
	}
	response << "}";
    }

    //----------------------------------  Send the result.
    debug_msg(dlvl_detail, trans->ident(), "Send source list string");
    xfer_string(trans, response.str());
    trans->set_state(transaction::ws_end_data);
}

//======================================  Thread initialization.
void
write_stage::init_thread(void) {
    sigignore(SIGPIPE);
}
